# -*- coding: utf-8 -*-

"""Control parser.

This module handles parsing control statement, which add annotations and namespaces to the document.

.. see also::

    https://wiki.openbel.org/display/BLD/Control+Records
"""

import logging
from typing import Any, Dict, List, Mapping, Optional, Pattern, Set

from pyparsing import And, Keyword, MatchFirst, ParseResults, Suppress, oneOf
from pyparsing import pyparsing_common as ppc

from .baseparser import BaseParser
from .utils import delimited_quoted_list, delimited_unquoted_list, is_int, qid, quote
from .. import constants as pc
from ..constants import (
    ANNOTATIONS,
    BEL_KEYWORD_ALL,
    BEL_KEYWORD_CITATION,
    BEL_KEYWORD_EVIDENCE,
    BEL_KEYWORD_SET,
    BEL_KEYWORD_STATEMENT_GROUP,
    BEL_KEYWORD_SUPPORT,
    BEL_KEYWORD_UNSET,
    CITATION,
    CITATION_TYPES,
    EVIDENCE,
)
from ..exceptions import (
    CitationTooLongException,
    CitationTooShortException,
    IllegalAnnotationValueWarning,
    InvalidCitationType,
    InvalidPubMedIdentifierWarning,
    MissingAnnotationKeyWarning,
    MissingAnnotationRegexWarning,
    MissingCitationException,
    UndefinedAnnotationWarning,
)
from ..language import CitationDict, Entity

__all__ = ["ControlParser"]

logger = logging.getLogger(__name__)

set_tag = Keyword(BEL_KEYWORD_SET)
unset_tag = Keyword(BEL_KEYWORD_UNSET)
unset_all = Suppress(BEL_KEYWORD_ALL)

supporting_text_tags = oneOf([BEL_KEYWORD_EVIDENCE, BEL_KEYWORD_SUPPORT])

set_statement_group_stub = And([Suppress(BEL_KEYWORD_STATEMENT_GROUP), Suppress("="), qid("group")])
set_citation_stub = And([Suppress(BEL_KEYWORD_CITATION), Suppress("="), delimited_quoted_list("values")])
set_evidence_stub = And([Suppress(supporting_text_tags), Suppress("="), quote("value")])


class ControlParser(BaseParser):
    """A parser for BEL control statements.

    .. seealso::

        BEL 1.0 specification on `control records
        <http://openbel.org/language/version_1.0/bel_specification_version_1.0.html#_control_records>`_
    """

    def __init__(
        self,
        annotation_to_term: Optional[Mapping[str, Set[str]]] = None,
        annotation_to_pattern: Optional[Mapping[str, Pattern]] = None,
        annotation_to_local: Optional[Mapping[str, Set[str]]] = None,
        citation_clearing: bool = True,
        required_annotations: Optional[List[str]] = None,
    ) -> None:
        """Initialize the control statement parser.

        :param annotation_to_term: A dictionary of {annotation: set of valid values} defined with URL for parsing
        :param annotation_to_pattern: A dictionary of {annotation: regular expression string}
        :param annotation_to_local: A dictionary of {annotation: set of valid values} for parsing defined with LIST
        :param citation_clearing: Should :code:`SET Citation` statements clear evidence and all annotations?
        :param required_annotations: Annotations that are required
        """
        self.citation_clearing = citation_clearing

        self.annotation_to_term = annotation_to_term or {}
        self.annotation_to_pattern = annotation_to_pattern or {}
        self.annotation_to_local = annotation_to_local or {}

        self.statement_group = None
        self.citation_db = None
        self.citation_db_id = None
        self.evidence = None
        self.annotations = {}
        self.required_annotations = required_annotations or []

        annotation_key = ppc.identifier("key").setParseAction(self.handle_annotation_key)

        self.set_statement_group = set_statement_group_stub().setParseAction(self.handle_set_statement_group)
        self.set_citation = set_citation_stub.setParseAction(self.handle_set_citation)
        self.set_evidence = set_evidence_stub.setParseAction(self.handle_set_evidence)

        set_command_prefix = And([annotation_key("key"), Suppress("=")])
        self.set_command = set_command_prefix + qid("value")
        self.set_command.setParseAction(self.handle_set_command)

        self.set_command_list = set_command_prefix + delimited_quoted_list("values")
        self.set_command_list.setParseAction(self.handle_set_command_list)

        self.unset_command = annotation_key("key")
        self.unset_command.addParseAction(self.handle_unset_command)

        self.unset_evidence = supporting_text_tags(EVIDENCE)
        self.unset_evidence.setParseAction(self.handle_unset_evidence)

        self.unset_citation = Suppress(BEL_KEYWORD_CITATION)
        self.unset_citation.setParseAction(self.handle_unset_citation)

        self.unset_statement_group = Suppress(BEL_KEYWORD_STATEMENT_GROUP)
        self.unset_statement_group.setParseAction(self.handle_unset_statement_group)

        self.unset_list = delimited_unquoted_list("values")
        self.unset_list.setParseAction(self.handle_unset_list)

        self.unset_all = unset_all.setParseAction(self.handle_unset_all)

        self.set_statements = set_tag("action") + MatchFirst(
            [
                self.set_statement_group,
                self.set_citation,
                self.set_evidence,
                self.set_command,
                self.set_command_list,
            ]
        )

        self.unset_statements = unset_tag("action") + MatchFirst(
            [
                self.unset_all,
                self.unset_citation,
                self.unset_evidence,
                self.unset_statement_group,
                self.unset_command,
                self.unset_list,
            ]
        )

        self.language = self.set_statements | self.unset_statements

        super(ControlParser, self).__init__(self.language)

    @property
    def _in_debug_mode(self) -> bool:
        return not self.annotation_to_term and not self.annotation_to_pattern

    @property
    def citation_is_set(self) -> bool:
        """Check if the citation is set."""
        return self.citation_db is not None and self.citation_db_id is not None

    def has_enumerated_annotation(self, annotation: str) -> bool:
        """Check if the annotation is defined as an enumeration."""
        return annotation in self.annotation_to_term

    def has_regex_annotation(self, annotation: str) -> bool:
        """Check if the annotation is defined as a regular expression."""
        return annotation in self.annotation_to_pattern

    def has_local_annotation(self, annotation: str) -> bool:
        """Check if the annotation is defined locally."""
        return annotation in self.annotation_to_local

    def has_annotation(self, annotation: str) -> bool:
        """Check if the annotation is defined."""
        return (
            self.has_enumerated_annotation(annotation)
            or self.has_regex_annotation(annotation)
            or self.has_local_annotation(annotation)
        )

    def raise_for_undefined_annotation(self, line: str, position: int, annotation: str) -> None:
        """Raise an exception if the annotation is not defined.

        :raises: UndefinedAnnotationWarning
        """
        if self._in_debug_mode:
            return

        if not self.has_annotation(annotation):
            raise UndefinedAnnotationWarning(self.get_line_number(), line, position, annotation)

    def raise_for_invalid_annotation_value(self, line: str, position: int, key: str, value: str) -> None:
        """Raise an exception if the annotation is not defined.

        :raises: IllegalAnnotationValueWarning or MissingAnnotationRegexWarning
        """
        if self._in_debug_mode:
            return

        if self.has_enumerated_annotation(key) and value not in self.annotation_to_term[key]:
            raise IllegalAnnotationValueWarning(self.get_line_number(), line, position, key, value)

        elif self.has_regex_annotation(key) and not self.annotation_to_pattern[key].match(value):
            raise MissingAnnotationRegexWarning(self.get_line_number(), line, position, key, value)

        elif self.has_local_annotation(key) and value not in self.annotation_to_local[key]:  # TODO condense
            raise IllegalAnnotationValueWarning(self.get_line_number(), line, position, key, value)

    def raise_for_missing_citation(self, line: str, position: int) -> None:
        """Raise an exception if there is no citation present in the parser.

        :raises: MissingCitationException
        """
        if self.citation_clearing and not self.citation_is_set:
            raise MissingCitationException(self.get_line_number(), line, position)

    def handle_annotation_key(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle an annotation key before parsing to validate that it's either enumerated or as a regex.

        :raise: MissingCitationException or UndefinedAnnotationWarning
        """
        key = tokens["key"]
        self.raise_for_missing_citation(line, position)
        self.raise_for_undefined_annotation(line, position, key)
        return tokens

    def handle_set_statement_group(self, _, __, tokens: ParseResults) -> ParseResults:
        """Handle a ``SET STATEMENT_GROUP = "X"`` statement."""
        self.statement_group = tokens["group"]
        return tokens

    def handle_set_citation(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle a ``SET Citation = {"X", "Y", "Z", ...}`` statement."""
        self.clear_citation()

        values = tokens["values"]

        if len(values) < 2:
            raise CitationTooShortException(self.get_line_number(), line, position)

        citation_namespace = values[0].lower()
        citation_namespace = pc.CITATION_NORMALIZER.get(citation_namespace, citation_namespace)
        if citation_namespace not in CITATION_TYPES:
            raise InvalidCitationType(self.get_line_number(), line, position, citation_namespace)

        if 2 == len(values):
            citation_db_id = values[1]

        elif 6 < len(values):
            raise CitationTooLongException(self.get_line_number(), line, position)

        else:
            if 3 == len(values):
                logger.debug("Throwing away JOURNAL entry in position 2")
            else:
                logger.debug("Throwing away JOURNAL entry in position 2 and everything after position 3")

            citation_db_id = values[2]

        if citation_namespace == "pubmed" and not is_int(citation_db_id):
            raise InvalidPubMedIdentifierWarning(self.get_line_number(), line, position, citation_db_id)

        self.citation_db = citation_namespace
        self.citation_db_id = citation_db_id
        return tokens

    def handle_set_evidence(self, _, __, tokens: ParseResults) -> ParseResults:
        """Handle a ``SET Evidence = ""`` statement."""
        self.evidence = tokens["value"]
        return tokens

    def handle_set_command(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle a ``SET X = "Y"`` statement."""
        key, value = tokens["key"], tokens["value"]
        self.raise_for_invalid_annotation_value(line, position, key, value)
        self.annotations[key] = [Entity(namespace=key, identifier=value)]
        return tokens

    def handle_set_command_list(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle a ``SET X = {"Y", "Z", ...}`` statement."""
        key, values = tokens["key"], tokens["values"]
        for value in values:
            self.raise_for_invalid_annotation_value(line, position, key, value)
        self.annotations[key] = [Entity(namespace=key, identifier=value) for value in values]
        return tokens

    def handle_unset_statement_group(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Unset the statement group, or raises an exception if it is not set.

        :raises: MissingAnnotationKeyWarning
        """
        if self.statement_group is None:
            raise MissingAnnotationKeyWarning(self.get_line_number(), line, position, BEL_KEYWORD_STATEMENT_GROUP)
        self.statement_group = None
        return tokens

    def handle_unset_citation(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Unset the citation, or raise an exception if it is not set.

        :raises: MissingAnnotationKeyWarning
        """
        if not self.citation_is_set:
            raise MissingAnnotationKeyWarning(self.get_line_number(), line, position, BEL_KEYWORD_CITATION)

        self.clear_citation()
        return tokens

    def handle_unset_evidence(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Unset the evidence, or throws an exception if it is not already set.

        The value for ``tokens[EVIDENCE]`` corresponds to which alternate of SupportingText or Evidence was used in
        the BEL script.

        :raises: MissingAnnotationKeyWarning
        """
        if self.evidence is None:
            raise MissingAnnotationKeyWarning(self.get_line_number(), line, position, tokens[EVIDENCE])
        self.evidence = None
        return tokens

    def validate_unset_command(self, line: str, position: int, annotation: str) -> None:
        """Raise an exception when trying to ``UNSET X`` if ``X`` is not already set.

        :raises: MissingAnnotationKeyWarning
        """
        if annotation not in self.annotations:
            raise MissingAnnotationKeyWarning(self.get_line_number(), line, position, annotation)

    def handle_unset_command(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle an ``UNSET X`` statement or raises an exception if it is not already set.

        :raises: MissingAnnotationKeyWarning
        """
        key = tokens["key"]
        self.validate_unset_command(line, position, key)
        del self.annotations[key]
        return tokens

    def handle_unset_list(self, line: str, position: int, tokens: ParseResults) -> ParseResults:
        """Handle ``UNSET {A, B, ...}`` or raises an exception of any of them are not present.

        Consider that all unsets are in peril if just one of them is wrong!

        :raises: MissingAnnotationKeyWarning
        """
        for key in tokens["values"]:
            if key in {BEL_KEYWORD_EVIDENCE, BEL_KEYWORD_SUPPORT}:
                self.evidence = None
            else:
                self.validate_unset_command(line, position, key)
                del self.annotations[key]

        return tokens

    def handle_unset_all(self, _, __, tokens) -> ParseResults:
        """Handle an ``UNSET_ALL`` statement."""
        self.clear()
        return tokens

    def get_annotations(self) -> Dict[str, Any]:
        """Get the current annotations."""
        return {
            EVIDENCE: self.evidence,
            CITATION: self.get_citation(),
            ANNOTATIONS: self.annotations.copy(),
        }

    def get_citation(self) -> Optional[CitationDict]:
        """Get the citation dictionary."""
        return (
            CitationDict(namespace=self.citation_db, identifier=self.citation_db_id)
            if self.citation_db and self.citation_db_id
            else None
        )

    def get_missing_required_annotations(self) -> List[str]:
        """Return missing required annotations."""
        return [
            required_annotation
            for required_annotation in self.required_annotations
            if required_annotation not in self.annotations
        ]

    def clear_citation(self) -> None:
        """Clear the citation and if citation clearing is enabled, clear the evidence and annotations."""
        self.citation_db = None
        self.citation_db_id = None

        if self.citation_clearing:
            self.evidence = None
            self.annotations.clear()

    def clear(self) -> None:
        """Clear the statement_group, citation, evidence, and annotations."""
        self.statement_group = None
        self.citation_db = None
        self.citation_db_id = None
        self.evidence = None
        self.annotations.clear()



# -*- coding: utf-8 -*-

"""Tests for manager functions handling BEL networks."""

import time
import unittest
from collections import Counter
from random import randint

from pybel import BELGraph, from_bel_script, from_database, to_database
from pybel.constants import (
    ABUNDANCE,
    BIOPROCESS,
    CITATION_TYPE_PUBMED,
    DECREASES,
    HAS_PRODUCT,
    HAS_REACTANT,
    IDENTIFIER,
    INCREASES,
    METADATA_NAME,
    METADATA_VERSION,
    MIRNA,
    NAMESPACE,
    PART_OF,
    PATHOLOGY,
    PROTEIN,
    RELATION,
)
from pybel.dsl import (
    BaseEntity,
    ComplexAbundance,
    CompositeAbundance,
    EnumeratedFusionRange,
    Fragment,
    Gene,
    GeneFusion,
    GeneModification,
    Hgvs,
    NamedComplexAbundance,
    Pathology,
    Protein,
    ProteinModification,
    Reaction,
    activity,
    degradation,
    location,
    secretion,
    translocation,
)
from pybel.dsl.namespaces import chebi, hgnc, mirbase
from pybel.examples import ras_tloc_graph, sialic_acid_graph
from pybel.language import Entity
from pybel.manager import models
from pybel.manager.models import Author, Citation, Edge, Evidence, NamespaceEntry, Node
from pybel.testing.cases import (
    FleetingTemporaryCacheMixin,
    TemporaryCacheClsMixin,
    TemporaryCacheMixin,
)
from pybel.testing.constants import test_bel_simple
from pybel.testing.mocks import mock_bel_resources
from pybel.testing.utils import make_dummy_annotations, make_dummy_namespaces, n
from tests.constants import (
    BelReconstitutionMixin,
    akt1,
    casp8,
    egfr,
    expected_test_simple_metadata,
    fadd,
    test_citation_dict,
    test_evidence_text,
)

fos = hgnc(name="FOS")
jun = hgnc(name="JUN")
mirna_1 = mirbase(name=n())
mirna_2 = mirbase(name=n())
pathology_1 = Pathology("DO", n())
ap1_complex = ComplexAbundance([fos, jun])

egfr_dimer = ComplexAbundance([egfr, egfr])

yfg_data = hgnc(name="YFG")
e2f4_data = hgnc(name="E2F4")
bound_ap1_e2f4 = ComplexAbundance([ap1_complex, e2f4_data])

superoxide = chebi(name="superoxide")
hydrogen_peroxide = chebi(name="hydrogen peroxide")
oxygen = chebi(name="oxygen")
superoxide_decomposition = Reaction(reactants=[superoxide], products=[hydrogen_peroxide, oxygen])


def assert_unqualified_edge(test_case, u: BaseEntity, v: BaseEntity, rel: str) -> None:
    """Assert there's only one edge and get the data for it"""
    test_case.assertIn(u, test_case.graph)
    test_case.assertIn(v, test_case.graph[u])
    edges = list(test_case.graph[u][v].values())
    test_case.assertEqual(1, len(edges))
    data = edges[0]
    test_case.assertEqual(rel, data[RELATION])


class TestNetworkCache(BelReconstitutionMixin, FleetingTemporaryCacheMixin):
    def test_get_network_missing(self):
        network = self.manager.get_most_recent_network_by_name("This network is not here")
        self.assertIsNone(network)

    def test_get_graph_missing(self):
        network = self.manager.get_graph_by_most_recent("This network is not here")
        self.assertIsNone(network)

    @mock_bel_resources
    def test_reload(self, mock_get):
        """Test that a graph with the same name and version can't be added twice."""
        graph = sialic_acid_graph.copy()
        self.assertEqual("1.0.0", graph.version)

        to_database(graph, manager=self.manager)
        time.sleep(1)

        self.assertEqual(1, self.manager.count_networks())

        networks = self.manager.list_networks()
        self.assertEqual(1, len(networks))

        network = networks[0]
        self.assertEqual(graph.name, network.name)
        self.assertEqual(graph.version, network.version)
        self.assertEqual(graph.description, network.description)

        reconstituted = self.manager.get_graph_by_name_version(graph.name, graph.version)

        self.assertIsInstance(reconstituted, BELGraph)
        self.assertEqual(graph.nodes(data=True), reconstituted.nodes(data=True))
        # self.bel_thorough_reconstituted(reconstituted)

        self.assertEqual(1, self.manager.count_networks())

        graph_copy = graph.copy()
        graph_copy.version = "1.0.1"
        network_copy = self.manager.insert_graph(graph_copy)

        time.sleep(1)  # Sleep so the first graph always definitely goes in first

        self.assertNotEqual(network.id, network_copy.id)

        self.assertTrue(self.manager.has_name_version(graph_copy.name, graph_copy.version))
        self.assertFalse(self.manager.has_name_version("wrong name", "0.1.2"))
        self.assertFalse(self.manager.has_name_version(graph_copy.name, "0.1.2"))
        self.assertFalse(self.manager.has_name_version("wrong name", graph_copy.version))

        self.assertEqual(2, self.manager.count_networks())

        self.assertEqual("1.0.1", self.manager.get_most_recent_network_by_name(graph.name).version)

        query_ids = {-1, network.id, network_copy.id}
        query_networks_result = self.manager.get_networks_by_ids(query_ids)
        self.assertEqual(2, len(query_networks_result))
        self.assertEqual(
            {network.id, network_copy.id},
            {network.id for network in query_networks_result},
        )

        expected_versions = {"1.0.1", "1.0.0"}
        self.assertEqual(expected_versions, set(self.manager.get_network_versions(graph.name)))

        exact_name_version = from_database(graph.name, graph.version, manager=self.manager)
        self.assertEqual(graph.name, exact_name_version.name)
        self.assertEqual(graph.version, exact_name_version.version)

        exact_name_version = from_database(graph.name, "1.0.1", manager=self.manager)
        self.assertEqual(graph.name, exact_name_version.name)
        self.assertEqual("1.0.1", exact_name_version.version)

        most_recent_version = from_database(graph.name, manager=self.manager)
        self.assertEqual(graph.name, most_recent_version.name)
        self.assertEqual("1.0.1", exact_name_version.version)

        recent_networks = list(self.manager.list_recent_networks())  # just try it to see if it fails
        self.assertIsNotNone(recent_networks)
        self.assertEqual([(network.name, "1.0.1")], [(n.name, n.version) for n in recent_networks])
        self.assertEqual("1.0.1", recent_networks[0].version)

    @mock_bel_resources
    def test_upload_with_tloc(self, mock_get):
        """Test that the RAS translocation example graph can be uploaded."""
        make_dummy_namespaces(self.manager, ras_tloc_graph)
        to_database(ras_tloc_graph, manager=self.manager)


class TestTemporaryInsertNetwork(TemporaryCacheMixin):
    def test_insert_with_list_annotations(self):
        """This test checks that graphs that contain list annotations, which aren't cached, can be loaded properly
        into the database."""
        graph = BELGraph(name="test", version="0.0.0")
        graph.annotation_list["TEST"] = {"a", "b", "c"}

        graph.add_increases(
            fos,
            jun,
            evidence=test_evidence_text,
            citation=test_citation_dict,
            annotations={"TEST": "a"},
        )

        make_dummy_namespaces(self.manager, graph)

        with mock_bel_resources:
            self.manager.insert_graph(graph)

        # TODO check that the database doesn't have anything for TEST in it


class TestTypedQuery(TemporaryCacheMixin):
    def setUp(self):
        super().setUp()

        graph = BELGraph(name="test", version="0.0.0")
        graph.annotation_list["TEST"] = {"a", "b", "c"}

        graph.add_positive_correlation(
            mirna_1,
            pathology_1,
            evidence=n(),
            citation=n(),
        )

        graph.add_negative_correlation(
            mirna_2,
            pathology_1,
            evidence=n(),
            citation=n(),
        )

        make_dummy_namespaces(self.manager, graph)
        make_dummy_annotations(self.manager, graph)

        with mock_bel_resources:
            self.manager.insert_graph(graph)

    def test_query_edge_source_type(self):
        rv = self.manager.query_edges(source_function=MIRNA).all()
        self.assertEqual(2, len(rv))

        rv = self.manager.query_edges(target_function=PATHOLOGY).all()
        self.assertEqual(2, len(rv))


class TestQuery(TemporaryCacheMixin):
    def setUp(self):
        super(TestQuery, self).setUp()

        graph = BELGraph(name="test", version="0.0.0")
        graph.annotation_list["TEST"] = {"a", "b", "c"}

        graph.add_increases(
            fos,
            jun,
            evidence=test_evidence_text,
            citation=test_citation_dict,
            annotations={"TEST": "a"},
        )

        make_dummy_namespaces(self.manager, graph)
        make_dummy_annotations(self.manager, graph)

        with mock_bel_resources:
            self.manager.insert_graph(graph)

    def test_query_node_bel_1(self):
        rv = self.manager.query_nodes(bel="p(HGNC:FOS)").all()
        self.assertEqual(1, len(rv))
        self.assertEqual(fos, rv[0].to_json())

    def test_query_node_bel_2(self):
        rv = self.manager.query_nodes(bel="p(HGNC:JUN)").all()
        self.assertEqual(1, len(rv))
        self.assertEqual(jun, rv[0].to_json())

    def test_query_node_namespace_wildcard(self):
        rv = self.manager.query_nodes(namespace="HG%").all()
        self.assertEqual(2, len(rv))
        self.assertTrue(any(x.to_json() == fos for x in rv))
        self.assertTrue(any(x.to_json() == jun for x in rv))

    def test_query_node_name_wildcard(self):
        rv = self.manager.query_nodes(name="%J%").all()
        self.assertEqual(1, len(rv), 1)
        self.assertEqual(jun, rv[0].to_json())

    def test_query_node_type(self):
        rv = self.manager.query_nodes(type=PROTEIN).all()
        self.assertEqual(2, len(rv))

    def test_query_node_type_missing(self):
        rv = self.manager.query_nodes(type=ABUNDANCE).all()
        self.assertEqual(0, len(rv))

    def test_query_edge_by_bel(self):
        rv = self.manager.query_edges(bel="p(HGNC:FOS) increases p(HGNC:JUN)").all()
        self.assertEqual(1, len(rv))

    def test_query_edge_by_relation_wildcard(self):
        # relation like, data
        increased_list = self.manager.query_edges(relation="increase%").all()
        self.assertEqual(1, len(increased_list))
        # self.assertIn(..., increased_list)

    def test_query_edge_by_evidence_wildcard(self):
        # evidence like, data
        evidence_list = self.manager.search_edges_with_evidence(evidence="%3%")
        self.assertEqual(len(evidence_list), 0)

        evidence_list = self.manager.search_edges_with_evidence(evidence="%Twit%")
        self.assertEqual(len(evidence_list), 1)

    def test_query_edge_by_mixed_no_result(self):
        # no result
        # FIXME what should this return
        empty_list = self.manager.query_edges(source="p(HGNC:FADD)", relation=DECREASES)
        self.assertEqual(len(empty_list), 0)

    def test_query_edge_by_mixed(self):
        # source, relation, data
        source_list = self.manager.query_edges(source="p(HGNC:FOS)", relation=INCREASES).all()
        self.assertEqual(len(source_list), 1)

    def test_query_edge_by_source_function(self):
        edges = self.manager.query_edges(source_function=PROTEIN).all()
        self.assertEqual(1, len(edges), msg="Wrong number of edges: {}".format(edges))

        edges = self.manager.query_edges(source_function=BIOPROCESS).all()
        self.assertEqual(0, len(edges), msg="Wrong number of edges: {}".format(edges))

    def test_query_edge_by_target_function(self):
        edges = self.manager.query_edges(target_function=PROTEIN).all()
        self.assertEqual(1, len(edges), msg="Wrong number of edges: {}".format(edges))

        edges = self.manager.query_edges(target_function=PATHOLOGY).all()
        self.assertEqual(0, len(edges), msg="Wrong number of edges: {}".format(edges))

    def test_query_citation_by_type(self):
        rv = self.manager.query_citations(db=CITATION_TYPE_PUBMED)
        self.assertEqual(1, len(rv))
        self.assertTrue(rv[0].is_pubmed)
        self.assertFalse(rv[0].is_enriched)

    def test_query_citaiton_by_reference(self):
        rv = self.manager.query_citations(db=CITATION_TYPE_PUBMED, db_id=test_citation_dict[IDENTIFIER])
        self.assertEqual(1, len(rv))
        self.assertTrue(rv[0].is_pubmed)
        self.assertFalse(rv[0].is_enriched)
        self.assertEqual(test_citation_dict, rv[0].to_json())

    @unittest.skip
    def test_query_by_author_wildcard(self):
        author_list = self.manager.query_citations(author="Example%")
        self.assertEqual(len(author_list), 1)

    @unittest.skip
    def test_query_by_name(self):
        # type, name, data
        name_dict_list = self.manager.query_citations(
            db=CITATION_TYPE_PUBMED,
            name="That other article from last week",
        )
        self.assertEqual(len(name_dict_list), 1)
        # self.assertIn(..., name_dict_list)

    @unittest.skip
    def test_query_by_name_wildcard(self):
        # type, name like, data
        name_dict_list2 = self.manager.query_citations(db=CITATION_TYPE_PUBMED, name="%article from%")
        self.assertEqual(len(name_dict_list2), 2)
        # self.assertIn(..., name_dict_list2)
        # self.assertIn(..., name_dict_list2)


class TestEnsure(TemporaryCacheMixin):
    def test_get_or_create_citation(self):
        reference = str(randint(1, 1000000))
        citation_dict = {
            NAMESPACE: CITATION_TYPE_PUBMED,
            IDENTIFIER: reference,
        }

        citation = self.manager.get_or_create_citation(**citation_dict)
        self.manager.session.commit()

        self.assertIsInstance(citation, Citation)
        self.assertEqual(citation_dict, citation.to_json())

        citation_reloaded_from_reference = self.manager.get_citation_by_pmid(reference)
        self.assertIsNotNone(citation_reloaded_from_reference)
        self.assertEqual(citation_dict, citation_reloaded_from_reference.to_json())

        citation_reloaded_from_dict = self.manager.get_or_create_citation(**citation_dict)
        self.assertIsNotNone(citation_reloaded_from_dict)
        self.assertEqual(citation_dict, citation_reloaded_from_dict.to_json())

        citation_reloaded_from_reference = self.manager.get_citation_by_reference(
            citation_dict[NAMESPACE],
            citation_dict[IDENTIFIER],
        )
        self.assertIsNotNone(citation_reloaded_from_reference)
        self.assertEqual(citation_dict, citation_reloaded_from_reference.to_json())

    def test_get_or_create_evidence(self):
        citation_db, citation_ref = CITATION_TYPE_PUBMED, str(randint(1, 1000000))
        basic_citation = self.manager.get_or_create_citation(namespace=citation_db, identifier=citation_ref)
        utf8_test_evidence = "Yes, all the information is true! This contains a unicode alpha: α"

        evidence = self.manager.get_or_create_evidence(basic_citation, utf8_test_evidence)
        self.assertIsInstance(evidence, Evidence)
        self.assertIn(evidence, self.manager.object_cache_evidence.values())

        # Objects cached?
        reloaded_evidence = self.manager.get_or_create_evidence(basic_citation, utf8_test_evidence)
        self.assertEqual(evidence, reloaded_evidence)

    def test_get_or_create_author(self):
        """This tests getting or creating author with unicode characters"""
        author_name = "Jαckson M"

        # Create
        author = self.manager.get_or_create_author(author_name)
        self.manager.session.commit()

        self.assertIsInstance(author, Author)
        self.assertEqual(author.name, author_name)

        author_from_name = self.manager.get_author_by_name(author_name)
        self.assertIsNotNone(author_from_name)
        self.assertEqual(author_name, author_from_name.name)

        # Get
        author_from_get = self.manager.get_or_create_author(author_name)
        self.assertEqual(author.name, author_from_get.name)
        self.assertEqual(author, author_from_get)


class TestEdgeStore(TemporaryCacheClsMixin, BelReconstitutionMixin):
    """Tests that the cache can be queried."""

    @classmethod
    def setUpClass(cls):
        """Set up the class with a BEL graph and its corresponding SQLAlchemy model."""
        super().setUpClass()

        with mock_bel_resources:
            cls.graph = from_bel_script(test_bel_simple, manager=cls.manager, disallow_nested=False)
            cls.network = cls.manager.insert_graph(cls.graph)

    def test_citations(self):
        citations = self.manager.session.query(Citation).all()
        self.assertEqual(2, len(citations), msg="Citations: {}".format(citations))

        citation_db_ids = {"123455", "123456"}
        self.assertEqual(citation_db_ids, {citation.db_id for citation in citations})

    def test_evidences(self):
        evidences = self.manager.session.query(Evidence).all()
        self.assertEqual(3, len(evidences))

        evidences_texts = {"Evidence 1 w extra notes", "Evidence 2", "Evidence 3"}
        self.assertEqual(evidences_texts, {evidence.text for evidence in evidences})

    def test_nodes(self):
        nodes = self.manager.session.query(Node).all()
        self.assertEqual(4, len(nodes))

    def test_edges(self):
        edges = self.manager.session.query(Edge).all()

        x = Counter((e.source.bel, e.target.bel) for e in edges)

        d = {
            (akt1.as_bel(), egfr.as_bel()): 1,
            (egfr.as_bel(), fadd.as_bel()): 1,
            (egfr.as_bel(), casp8.as_bel()): 1,
            (fadd.as_bel(), casp8.as_bel()): 1,
            (akt1.as_bel(), casp8.as_bel()): 1,  # two way association
            (casp8.as_bel(), akt1.as_bel()): 1,  # two way association
        }

        self.assertEqual(dict(x), d)

        network_edge_associations = (
            self.manager.session.query(models.network_edge).filter_by(network_id=self.network.id).all()
        )

        self.assertEqual(
            {network_edge_association.edge_id for network_edge_association in network_edge_associations},
            {edge.id for edge in edges},
        )

    def test_reconstitute(self):
        g2 = self.manager.get_graph_by_name_version(
            expected_test_simple_metadata[METADATA_NAME],
            expected_test_simple_metadata[METADATA_VERSION],
        )
        self.bel_simple_reconstituted(g2)


class TestAddNodeFromData(unittest.TestCase):
    def setUp(self):
        self.graph = BELGraph()

    def test_simple(self):
        self.graph.add_node_from_data(yfg_data)
        self.assertIn(yfg_data, self.graph)
        self.assertEqual(1, self.graph.number_of_nodes())

    def test_single_variant(self):
        node_data = Gene("HGNC", "AKT1", variants=Hgvs("p.Phe508del"))
        node_parent_data = node_data.get_parent()

        self.graph.add_node_from_data(node_data)
        self.assertIn(node_data, self.graph)
        self.assertIn(node_parent_data, self.graph)
        self.assertEqual(2, self.graph.number_of_nodes())
        self.assertEqual(1, self.graph.number_of_edges())

    def test_multiple_variants(self):
        node_data = Gene("HGNC", "AKT1", variants=[Hgvs("p.Phe508del"), Hgvs("p.Phe509del")])
        node_parent_data = node_data.get_parent()
        node_parent_tuple = node_parent_data

        self.graph.add_node_from_data(node_data)
        self.assertIn(node_data, self.graph)
        self.assertIn(node_parent_tuple, self.graph)
        self.assertEqual(2, self.graph.number_of_nodes())
        self.assertEqual(1, self.graph.number_of_edges())

    def test_fusion(self):
        node_data = GeneFusion(
            partner_5p=Gene("HGNC", "TMPRSS2"),
            partner_3p=Gene("HGNC", "ERG"),
            range_5p=EnumeratedFusionRange("c", 1, 79),
            range_3p=EnumeratedFusionRange("c", 312, 5034),
        )
        node_data = node_data

        self.graph.add_node_from_data(node_data)
        self.assertIn(node_data, self.graph)
        self.assertEqual(1, self.graph.number_of_nodes())
        self.assertEqual(0, self.graph.number_of_edges())

    def test_composite(self):
        il23 = NamedComplexAbundance(namespace="GO", name="interleukin-23 complex")
        il6 = Protein(namespace="HGNC", name="IL6")
        node_data = CompositeAbundance([il23, il6])

        self.graph.add_node_from_data(node_data)
        self.assertIn(node_data, self.graph)
        self.assertEqual(3, self.graph.number_of_nodes())
        self.assertIn(il6, self.graph, msg="Nodes:\n".format("\n".join(map(str, self.graph))))
        self.assertIn(il23, self.graph)
        self.assertEqual(2, self.graph.number_of_edges())

        self.assertIn(node_data, self.graph[il6])
        edges = list(self.graph[il6][node_data].values())
        self.assertEqual(1, len(edges))
        data = edges[0]
        self.assertEqual(PART_OF, data[RELATION])

        self.assertIn(node_data, self.graph[il23])
        edges = list(self.graph[il23][node_data].values())
        self.assertEqual(1, len(edges))
        data = edges[0]
        self.assertEqual(PART_OF, data[RELATION])

    def test_reaction(self):
        self.graph.add_node_from_data(superoxide_decomposition)
        self.assertIn(superoxide_decomposition, self.graph)
        self.assertEqual(4, self.graph.number_of_nodes())
        self.assertEqual(3, self.graph.number_of_edges())

        assert_unqualified_edge(self, superoxide_decomposition, superoxide, HAS_REACTANT)
        assert_unqualified_edge(self, superoxide_decomposition, hydrogen_peroxide, HAS_PRODUCT)
        assert_unqualified_edge(self, superoxide_decomposition, oxygen, HAS_PRODUCT)

    def test_complex(self):
        node = ComplexAbundance(members=[fos, jun])

        self.graph.add_node_from_data(node)
        self.assertIn(node, self.graph)
        self.assertEqual(3, self.graph.number_of_nodes())
        self.assertEqual(2, self.graph.number_of_edges())

        assert_unqualified_edge(self, fos, node, PART_OF)
        assert_unqualified_edge(self, jun, node, PART_OF)

    def test_dimer_complex(self):
        """Tests what happens if a BEL statement complex(p(X), p(X)) is added"""
        self.graph.add_node_from_data(egfr_dimer)

        self.assertIn(egfr, self.graph)
        self.assertIn(egfr_dimer, self.graph)
        self.assertEqual(2, self.graph.number_of_nodes())
        self.assertEqual(1, self.graph.number_of_edges())

        assert_unqualified_edge(self, egfr, egfr_dimer, PART_OF)

    def test_nested_complex(self):
        """Checks what happens if a theoretical BEL statement `complex(p(X), complex(p(Y), p(Z)))` is added"""
        self.graph.add_node_from_data(bound_ap1_e2f4)
        self.assertIn(bound_ap1_e2f4, self.graph)
        self.assertEqual(5, self.graph.number_of_nodes())
        self.assertIn(fos, self.graph)
        self.assertIn(jun, self.graph)
        self.assertIn(e2f4_data, self.graph)
        self.assertIn(ap1_complex, self.graph)
        self.assertEqual(4, self.graph.number_of_edges())

        assert_unqualified_edge(self, fos, ap1_complex, PART_OF)
        assert_unqualified_edge(self, jun, ap1_complex, PART_OF)
        assert_unqualified_edge(self, ap1_complex, bound_ap1_e2f4, PART_OF)
        assert_unqualified_edge(self, e2f4_data, bound_ap1_e2f4, PART_OF)


class TestReconstituteNodeTuples(TemporaryCacheMixin):
    """Tests the ability to go from PyBEL to relational database"""

    def _help_reconstitute(self, node: BaseEntity, number_nodes: int, number_edges: int):
        """Help test the round-trip conversion from PyBEL data dictionary to node model."""
        self.assertIsInstance(node, BaseEntity)

        graph = BELGraph(name="test", version="0.0.0")
        graph.add_node_from_data(node)

        make_dummy_namespaces(self.manager, graph)

        self.manager.insert_graph(graph)
        self.assertEqual(number_nodes, self.manager.count_nodes())
        self.assertEqual(number_edges, self.manager.count_edges())

        node_model = self.manager.get_or_create_node(graph, node)
        self.assertEqual(node.md5, node_model.md5)
        self.manager.session.commit()

        self.assertEqual(node, node_model.to_json())
        self.assertEqual(node, self.manager.get_dsl_by_hash(node.md5))

    @mock_bel_resources
    def test_simple(self, mock):
        self._help_reconstitute(yfg_data, 1, 0)

    @mock_bel_resources
    def test_Hgvs(self, mock):
        node_data = Gene(namespace="HGNC", name="AKT1", variants=Hgvs("p.Phe508del"))
        self._help_reconstitute(node_data, 2, 1)

    @mock_bel_resources
    def test_fragment_unspecified(self, mock):
        dummy_namespace = n()
        dummy_name = n()
        node_data = Protein(namespace=dummy_namespace, name=dummy_name, variants=[Fragment()])
        self._help_reconstitute(node_data, 2, 1)

    @mock_bel_resources
    def test_fragment_specified(self, mock):
        dummy_namespace = n()
        dummy_name = n()
        node_data = Protein(
            namespace=dummy_namespace,
            name=dummy_name,
            variants=[Fragment(start=5, stop=8)],
        )
        self._help_reconstitute(node_data, 2, 1)

    @mock_bel_resources
    def test_fragment_specified_start_only(self, mock):
        dummy_namespace = n()
        dummy_name = n()
        node_data = Protein(
            namespace=dummy_namespace,
            name=dummy_name,
            variants=[Fragment(start=5, stop="*")],
        )
        self._help_reconstitute(node_data, 2, 1)

    @mock_bel_resources
    def test_fragment_specified_end_only(self, mock):
        dummy_namespace = n()
        dummy_name = n()
        node_data = Protein(
            namespace=dummy_namespace,
            name=dummy_name,
            variants=[Fragment(start="*", stop=1000)],
        )
        self._help_reconstitute(node_data, 2, 1)

    @mock_bel_resources
    def test_gmod_custom(self, mock):
        """Tests a gene modification that uses a non-default namespace"""
        dummy_namespace = "HGNC"
        dummy_name = "AKT1"
        dummy_mod_namespace = "GO"
        dummy_mod_name = "DNA Methylation"

        node_data = Gene(
            namespace=dummy_namespace,
            name=dummy_name,
            variants=[GeneModification(name=dummy_mod_name, namespace=dummy_mod_namespace)],
        )
        self._help_reconstitute(node_data, 2, 1)

    @mock_bel_resources
    def test_gmod_default(self, mock):
        """Test a gene modification that uses the BEL default namespace."""
        dummy_namespace = n()
        dummy_name = n()

        node_data = Gene(
            namespace=dummy_namespace,
            name=dummy_name,
            variants=[GeneModification("Me")],
        )
        self._help_reconstitute(node_data, 2, 1)

    @mock_bel_resources
    def test_pmod_default_simple(self, mock):
        dummy_namespace = n()
        dummy_name = n()

        node_data = Protein(
            namespace=dummy_namespace,
            name=dummy_name,
            variants=[ProteinModification("Me")],
        )
        self._help_reconstitute(node_data, 2, 1)

    @mock_bel_resources
    def test_pmod_custom_simple(self, mock):
        dummy_namespace = "HGNC"
        dummy_name = "AKT1"
        dummy_mod_namespace = "GO"
        dummy_mod_name = "Protein phosphorylation"

        node_data = Protein(
            namespace=dummy_namespace,
            name=dummy_name,
            variants=[ProteinModification(name=dummy_mod_name, namespace=dummy_mod_namespace)],
        )
        self._help_reconstitute(node_data, 2, 1)

    @mock_bel_resources
    def test_pmod_default_with_residue(self, mock):
        dummy_namespace = n()
        dummy_name = n()

        node_data = Protein(
            namespace=dummy_namespace,
            name=dummy_name,
            variants=[ProteinModification("Me", code="Ser")],
        )
        self._help_reconstitute(node_data, 2, 1)

    @mock_bel_resources
    def test_pmod_custom_with_residue(self, mock):
        dummy_namespace = "HGNC"
        dummy_name = "AKT1"
        dummy_mod_namespace = "GO"
        dummy_mod_name = "Protein phosphorylation"

        node_data = Protein(
            namespace=dummy_namespace,
            name=dummy_name,
            variants=[ProteinModification(name=dummy_mod_name, namespace=dummy_mod_namespace, code="Ser")],
        )
        self._help_reconstitute(node_data, 2, 1)

    @mock_bel_resources
    def test_pmod_default_full(self, mock):
        dummy_namespace = n()
        dummy_name = n()

        node_data = Protein(
            namespace=dummy_namespace,
            name=dummy_name,
            variants=[ProteinModification("Me", code="Ser", position=5)],
        )
        self._help_reconstitute(node_data, 2, 1)

    @mock_bel_resources
    def test_pmod_custom_full(self, mock):
        dummy_namespace = "HGNC"
        dummy_name = "AKT1"
        dummy_mod_namespace = "GO"
        dummy_mod_name = "Protein phosphorylation"

        node_data = Protein(
            namespace=dummy_namespace,
            name=dummy_name,
            variants=[
                ProteinModification(
                    name=dummy_mod_name,
                    namespace=dummy_mod_namespace,
                    code="Ser",
                    position=5,
                )
            ],
        )
        self._help_reconstitute(node_data, 2, 1)

    @mock_bel_resources
    def test_multiple_variants(self, mock):
        node_data = Gene(
            namespace="HGNC",
            name="AKT1",
            variants=[Hgvs("p.Phe508del"), Hgvs("p.Phe509del")],
        )
        self._help_reconstitute(node_data, 2, 1)

    @mock_bel_resources
    def test_fusion_specified(self, mock):
        node_data = GeneFusion(
            Gene("HGNC", "TMPRSS2"),
            Gene("HGNC", "ERG"),
            EnumeratedFusionRange("c", 1, 79),
            EnumeratedFusionRange("c", 312, 5034),
        )
        self._help_reconstitute(node_data, 1, 0)

    @mock_bel_resources
    def test_fusion_unspecified(self, mock):
        node_data = GeneFusion(
            Gene("HGNC", "TMPRSS2"),
            Gene("HGNC", "ERG"),
        )
        self._help_reconstitute(node_data, 1, 0)

    @mock_bel_resources
    def test_composite(self, mock):
        interleukin_23_complex = NamedComplexAbundance("GO", "interleukin-23 complex")
        il6 = hgnc(name="IL6")
        interleukin_23_and_il6 = CompositeAbundance([interleukin_23_complex, il6])

        self._help_reconstitute(interleukin_23_and_il6, 3, 2)

    @mock_bel_resources
    def test_reaction(self, mock):
        self._help_reconstitute(superoxide_decomposition, 4, 3)

    @mock_bel_resources
    def test_complex(self, mock):
        self._help_reconstitute(ap1_complex, 3, 2)

    @mock_bel_resources
    def test_nested_complex(self, mock):
        self._help_reconstitute(bound_ap1_e2f4, 5, 4)


class TestReconstituteEdges(TemporaryCacheMixin):
    """This class tests that edges with varying properties can be added and extracted losslessly"""

    def setUp(self):
        """Creates a unit test with a manager and graph"""
        super().setUp()
        self.graph = BELGraph(name=n(), version=n())
        self.graph.annotation_pattern["Species"] = r"\d+"

    @mock_bel_resources
    def test_translocation_default(self, mock):
        """This test checks that a translocation gets in the database properly"""
        self.graph.add_increases(
            Protein(name="F2", namespace="HGNC"),
            Protein(name="EDN1", namespace="HGNC"),
            evidence="In endothelial cells, ET-1 secretion is detectable under basal conditions, whereas thrombin "
            "induces its secretion.",
            citation="10473669",
            source_modifier=secretion(),
        )

        make_dummy_namespaces(self.manager, self.graph)

        network = self.manager.insert_graph(self.graph)
        self.assertEqual(2, network.nodes.count(), msg="Missing one or both of the nodes.")
        self.assertEqual(1, network.edges.count(), msg="Missing the edge")

        # edge = network.edges.first()
        # self.assertEqual(2, edge.properties.count())

    @mock_bel_resources
    def test_subject_translocation_custom_to_loc(self, mock):
        self.graph.add_increases(
            Protein(name="F2", namespace="HGNC"),
            Protein(name="EDN1", namespace="HGNC"),
            evidence="In endothelial cells, ET-1 secretion is detectable under basal conditions, whereas thrombin induces its secretion.",
            citation="10473669",
            source_modifier=translocation(
                from_loc=Entity(namespace="TEST", name="A"),
                to_loc=Entity(namespace="GO", name="extracellular space"),
            ),
        )

        make_dummy_namespaces(self.manager, self.graph)

        network = self.manager.insert_graph(self.graph)
        self.assertEqual(2, network.nodes.count())
        self.assertEqual(1, network.edges.count())

        edge = network.edges.first()
        # self.assertEqual(2, edge.properties.count())

    @mock_bel_resources
    def test_subject_activity_default(self, mock):
        p1_name = n()
        p2_name = n()

        self.graph.add_increases(
            Protein(name=p1_name, namespace="HGNC"),
            Protein(name=p2_name, namespace="HGNC"),
            evidence=n(),
            citation=n(),
            source_modifier=activity("kin"),
        )

        make_dummy_namespaces(self.manager, self.graph)

        network = self.manager.insert_graph(self.graph)
        self.assertEqual(2, network.nodes.count(), msg="number of nodes")
        self.assertEqual(1, network.edges.count(), msg="number of edges")

        kin_list = self.manager.session.query(NamespaceEntry).filter(NamespaceEntry.name == "kinase activity").all()
        self.assertEqual(1, len(kin_list), msg="number of kinase NamespaceEntrys")

        kin = list(kin_list)[0]
        self.assertEqual("kinase activity", kin.name)

    @mock_bel_resources
    def test_subject_activity_custom(self, mock):
        p1_name = n()
        p2_name = n()
        dummy_activity_namespace = n()
        dummy_activity_name = n()

        self.graph.add_increases(
            Protein(name=p1_name, namespace="HGNC"),
            Protein(name=p2_name, namespace="HGNC"),
            evidence=n(),
            citation=n(),
            source_modifier=activity(name=dummy_activity_name, namespace=dummy_activity_namespace),
        )

        make_dummy_namespaces(self.manager, self.graph)

        network = self.manager.insert_graph(self.graph)
        self.assertEqual(2, network.nodes.count())
        self.assertEqual(1, network.edges.count())

        kin_list = self.manager.session.query(NamespaceEntry).filter(NamespaceEntry.name == dummy_activity_name).all()
        self.assertEqual(1, len(kin_list))

        kin = list(kin_list)[0]
        self.assertEqual(dummy_activity_name, kin.name)

    @mock_bel_resources
    def test_object_activity_default(self, mock):
        p1_name = n()
        p2_name = n()

        self.graph.add_increases(
            Protein(name=p1_name, namespace="HGNC"),
            Protein(name=p2_name, namespace="HGNC"),
            evidence=n(),
            citation=n(),
            target_modifier=activity("kin"),
        )

        make_dummy_namespaces(self.manager, self.graph)

        network = self.manager.insert_graph(self.graph)
        self.assertEqual(2, network.nodes.count())
        self.assertEqual(1, network.edges.count())

        kin_list = self.manager.session.query(NamespaceEntry).filter(NamespaceEntry.name == "kinase activity").all()
        self.assertEqual(1, len(kin_list))

        kin = list(kin_list)[0]
        self.assertEqual("kinase activity", kin.name)

    @mock_bel_resources
    def test_object_activity_custom(self, mock):
        p1_name = n()
        p2_name = n()
        dummy_activity_namespace = n()
        dummy_activity_name = n()

        self.graph.add_increases(
            Protein(name=p1_name, namespace="HGNC"),
            Protein(name=p2_name, namespace="HGNC"),
            evidence=n(),
            citation=n(),
            target_modifier=activity(name=dummy_activity_name, namespace=dummy_activity_namespace),
        )

        make_dummy_namespaces(self.manager, self.graph)

        network = self.manager.insert_graph(self.graph)
        self.assertEqual(2, network.nodes.count())
        self.assertEqual(1, network.edges.count())

        kin_list = self.manager.session.query(NamespaceEntry).filter(NamespaceEntry.name == dummy_activity_name).all()
        self.assertEqual(1, len(kin_list))

        kin = list(kin_list)[0]
        self.assertEqual(dummy_activity_name, kin.name)

    def test_subject_degradation(self):
        self.graph.add_increases(
            Protein(name="YFG", namespace="HGNC"),
            Protein(name="YFG2", namespace="HGNC"),
            evidence=n(),
            citation=n(),
            source_modifier=degradation(),
        )
        make_dummy_namespaces(self.manager, self.graph)

        network = self.manager.insert_graph(self.graph)

        self.assertEqual(2, network.nodes.count())
        self.assertEqual(1, network.edges.count())

        edge = network.edges.first()
        # self.assertEqual(1, edge.properties.count())

    def test_object_degradation(self):
        self.graph.add_increases(
            Protein(name="YFG", namespace="HGNC"),
            Protein(name="YFG2", namespace="HGNC"),
            evidence=n(),
            citation=n(),
            target_modifier=degradation(),
        )
        make_dummy_namespaces(self.manager, self.graph)

        network = self.manager.insert_graph(self.graph)

        self.assertEqual(2, network.nodes.count())
        self.assertEqual(1, network.edges.count())

        edge = network.edges.first()
        # self.assertEqual(1, edge.properties.count())

    def test_subject_location(self):
        self.graph.add_increases(
            Protein(name="YFG", namespace="HGNC"),
            Protein(name="YFG2", namespace="HGNC"),
            evidence=n(),
            citation=n(),
            source_modifier=location(Entity(namespace="GO", name="nucleus", identifier="0005634")),
        )
        make_dummy_namespaces(self.manager, self.graph)

        network = self.manager.insert_graph(self.graph)

        self.assertEqual(2, network.nodes.count())
        self.assertEqual(1, network.edges.count())

        edge = network.edges.first()
        # self.assertEqual(1, edge.properties.count())

    def test_mixed_1(self):
        """Test mixed having location and something else."""
        self.graph.add_increases(
            Protein(namespace="HGNC", name="CDC42"),
            Protein(namespace="HGNC", name="PAK2"),
            evidence="""Summary: PAK proteins, a family of serine/threonine p21-activating kinases, include PAK1, PAK2,
         PAK3 and PAK4. PAK proteins are critical effectors that link Rho GTPases to cytoskeleton reorganization
         and nuclear signaling. They serve as targets for the small GTP binding proteins Cdc42 and Rac and have
         been implicated in a wide range of biological activities. PAK4 interacts specifically with the GTP-bound
         form of Cdc42Hs and weakly activates the JNK family of MAP kinases. PAK4 is a mediator of filopodia
         formation and may play a role in the reorganization of the actin cytoskeleton. Multiple alternatively
         spliced transcript variants encoding distinct isoforms have been found for this gene.""",
            citation={
                NAMESPACE: "Online Resource",
                IDENTIFIER: "PAK4 Hs ENTREZ Gene Summary",
            },
            annotations={"Species": "9606"},
            source_modifier=activity("gtp"),
            target_modifier=activity("kin"),
        )

        make_dummy_namespaces(self.manager, self.graph)
        make_dummy_annotations(self.manager, self.graph)

        network = self.manager.insert_graph(self.graph)
        self.assertEqual(2, network.nodes.count())
        self.assertEqual(1, network.edges.count())

        edge = network.edges.first()
        # self.assertEqual(2, edge.properties.count())
        # FIXME

    def test_mixed_2(self):
        """Tests both subject and object activity with location information as well."""
        self.graph.add_directly_increases(
            Protein(namespace="HGNC", name="HDAC4"),
            Protein(namespace="HGNC", name="MEF2A"),
            citation="10487761",
            evidence=""""In the nucleus, HDAC4 associates with the myocyte enhancer factor MEF2A. Binding of HDAC4 to
        MEF2A results in the repression of MEF2A transcriptional activation, a function that requires the
        deacetylase domain of HDAC4.""",
            annotations={"Species": "9606"},
            source_modifier=activity("cat", location=Entity(namespace="GO", name="nucleus")),
            target_modifier=activity("tscript", location=Entity(namespace="GO", name="nucleus")),
        )

        make_dummy_namespaces(self.manager, self.graph)
        make_dummy_annotations(self.manager, self.graph)

        network = self.manager.insert_graph(self.graph)
        self.assertEqual(2, network.nodes.count())
        self.assertEqual(1, network.edges.count())

        edge = network.edges.first()
        # self.assertEqual(4, edge.properties.count())
        # self.assertEqual(2, edge.properties.filter(Property.is_subject).count())
        # self.assertEqual(2, edge.properties.filter(not_(Property.is_subject)).count())


class TestNoAddNode(TemporaryCacheMixin):
    """Tests scenarios where an instance of :class:`BELGraph` may contain edges that refer to uncached resources, and
    therefore should not be added to the edge store."""

    @mock_bel_resources
    def test_regex_lookup(self, mock):  # FIXME this test needs to be put somewhere else
        """Test that regular expression nodes get love too."""
        graph = BELGraph(
            name="Regular Expression Test Graph",
            description="Help test regular expression namespaces",
            version="1.0.0",
        )
        dbsnp = "dbSNP"
        DBSNP_PATTERN = "rs[0-9]+"
        graph.namespace_pattern[dbsnp] = DBSNP_PATTERN

        rs1234 = Gene(namespace=dbsnp, name="rs1234")
        rs1235 = Gene(namespace=dbsnp, name="rs1235")

        graph.add_node_from_data(rs1234)
        graph.add_node_from_data(rs1235)

        rs1234_hash = rs1234.md5
        rs1235_hash = rs1235.md5

        self.manager.insert_graph(graph)

        rs1234_lookup = self.manager.get_node_by_hash(rs1234_hash)
        self.assertIsNotNone(rs1234_lookup)
        self.assertEqual("Gene", rs1234_lookup.type)
        self.assertEqual("g(dbSNP:rs1234)", rs1234_lookup.bel)
        self.assertEqual(rs1234_hash, rs1234_lookup.md5)
        self.assertIsNotNone(rs1234_lookup.namespace_entry)
        self.assertEqual("rs1234", rs1234_lookup.namespace_entry.name)
        self.assertEqual("dbSNP", rs1234_lookup.namespace_entry.namespace.keyword)
        self.assertEqual(DBSNP_PATTERN, rs1234_lookup.namespace_entry.namespace.pattern)

        rs1235_lookup = self.manager.get_node_by_hash(rs1235_hash)
        self.assertIsNotNone(rs1235_lookup)
        self.assertEqual("Gene", rs1235_lookup.type)
        self.assertEqual("g(dbSNP:rs1235)", rs1235_lookup.bel)
        self.assertEqual(rs1235_hash, rs1235_lookup.md5)
        self.assertIsNotNone(rs1235_lookup.namespace_entry)
        self.assertEqual("rs1235", rs1235_lookup.namespace_entry.name)
        self.assertEqual("dbSNP", rs1235_lookup.namespace_entry.namespace.keyword)
        self.assertEqual(DBSNP_PATTERN, rs1235_lookup.namespace_entry.namespace.pattern)


class TestEquivalentNodes(unittest.TestCase):
    def test_direct_has_namespace(self):
        graph = BELGraph()

        n1 = Protein(namespace="HGNC", name="CD33", identifier="1659")
        n2 = Protein(namespace="NOPE", name="NOPE", identifier="NOPE")

        graph.add_increases(n1, n2, citation=n(), evidence=n())

        self.assertEqual({n1}, graph.get_equivalent_nodes(n1))

        self.assertTrue(graph.node_has_namespace(n1, "HGNC"))
        self.assertFalse(graph.node_has_namespace(n2, "HGNC"))

    def test_indirect_has_namespace(self):
        graph = BELGraph()

        a = Protein(namespace="HGNC", name="CD33")
        b = Protein(namespace="HGNCID", identifier="1659")

        graph.add_equivalence(a, b)

        self.assertEqual({a, b}, graph.get_equivalent_nodes(a))
        self.assertEqual({a, b}, graph.get_equivalent_nodes(b))

        self.assertTrue(graph.node_has_namespace(a, "HGNC"))
        self.assertTrue(graph.node_has_namespace(b, "HGNC"))

    def test_triangle_has_namespace(self):
        graph = BELGraph()

        a = Protein(namespace="A", name="CD33")
        b = Protein(namespace="B", identifier="1659")
        c = Protein(namespace="C", identifier="1659")
        d = Protein(namespace="HGNC", identifier="1659")

        graph.add_equivalence(a, b)
        graph.add_equivalence(b, c)
        graph.add_equivalence(c, a)
        graph.add_equivalence(c, d)

        self.assertEqual({a, b, c, d}, graph.get_equivalent_nodes(a))
        self.assertEqual({a, b, c, d}, graph.get_equivalent_nodes(b))
        self.assertEqual({a, b, c, d}, graph.get_equivalent_nodes(c))
        self.assertEqual({a, b, c, d}, graph.get_equivalent_nodes(d))

        self.assertTrue(graph.node_has_namespace(a, "HGNC"))
        self.assertTrue(graph.node_has_namespace(b, "HGNC"))
        self.assertTrue(graph.node_has_namespace(c, "HGNC"))
        self.assertTrue(graph.node_has_namespace(d, "HGNC"))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""Test parsing variants."""

import logging
import re
import unittest

from pybel.constants import (
    CONCEPT,
    FRAGMENT,
    FRAGMENT_DESCRIPTION,
    FRAGMENT_MISSING,
    FRAGMENT_START,
    FRAGMENT_STOP,
    FUSION_MISSING,
    FUSION_REFERENCE,
    FUSION_START,
    FUSION_STOP,
    GMOD,
    IDENTIFIER,
    KIND,
    LOCATION,
    NAME,
    NAMESPACE,
    PARTNER_3P,
    PARTNER_5P,
    PMOD,
    PMOD_CODE,
    PMOD_POSITION,
    RANGE_3P,
    RANGE_5P,
)
from pybel.dsl import GeneModification, Hgvs, ProteinModification
from pybel.language import Entity
from pybel.parser import ConceptParser
from pybel.parser.modifiers import (
    get_fragment_language,
    get_fusion_language,
    get_gene_modification_language,
    get_gene_substitution_language,
    get_hgvs_language,
    get_location_language,
    get_protein_modification_language,
    get_protein_substitution_language,
    get_truncation_language,
)

log = logging.getLogger(__name__)


class TestHGVSParser(unittest.TestCase):
    def setUp(self):
        self.parser = get_hgvs_language()

    def test_protein_del(self):
        statement = "variant(p.Phe508del)"
        expected = Hgvs("p.Phe508del")
        result = self.parser.parseString(statement)
        self.assertEqual(expected, result.asDict())

    def test_protein_del_quoted(self):
        statement = 'variant("p.Phe508del")'
        expected = Hgvs("p.Phe508del")
        result = self.parser.parseString(statement)
        self.assertEqual(expected, result.asDict())

    def test_protein_mut(self):
        statement = "var(p.Gly576Ala)"
        expected = Hgvs("p.Gly576Ala")
        result = self.parser.parseString(statement)
        self.assertEqual(expected, result.asDict())

    def test_unspecified(self):
        statement = "var(=)"
        expected = Hgvs("=")
        result = self.parser.parseString(statement)
        self.assertEqual(expected, result.asDict())

    def test_frameshift(self):
        statement = "variant(p.Thr1220Lysfs)"
        expected = Hgvs("p.Thr1220Lysfs")
        result = self.parser.parseString(statement)
        self.assertEqual(expected, result.asDict())

    def test_snp(self):
        statement = "var(c.1521_1523delCTT)"
        expected = Hgvs("c.1521_1523delCTT")
        result = self.parser.parseString(statement)
        self.assertEqual(expected, result.asDict())

    def test_chromosome_1(self):
        statement = "variant(g.117199646_117199648delCTT)"
        expected = Hgvs("g.117199646_117199648delCTT")
        result = self.parser.parseString(statement)
        self.assertEqual(expected, result.asDict())

    def test_chromosome_2(self):
        statement = "var(c.1521_1523delCTT)"
        expected = Hgvs("c.1521_1523delCTT")
        result = self.parser.parseString(statement)
        self.assertEqual(expected, result.asDict())

    def test_rna_del(self):
        statement = "var(r.1653_1655delcuu)"
        expected = Hgvs("r.1653_1655delcuu")
        result = self.parser.parseString(statement)
        self.assertEqual(expected, result.asDict())

    def test_protein_trunc_triple(self):
        statement = "var(p.Cys65*)"
        result = self.parser.parseString(statement)
        expected = Hgvs("p.Cys65*")
        self.assertEqual(expected, result.asDict())

    def test_protein_trunc_legacy(self):
        statement = "var(p.65*)"
        result = self.parser.parseString(statement)
        expected = Hgvs("p.65*")
        self.assertEqual(expected, result.asDict())


class TestPmod(unittest.TestCase):
    def setUp(self):
        identifier_parser = ConceptParser(
            namespace_to_pattern={
                "MOD": re.compile(".*"),
                "HGNC": re.compile(".*"),
            }
        )
        self.parser = get_protein_modification_language(
            concept_qualified=identifier_parser.identifier_qualified,
            concept_fqualified=identifier_parser.identifier_fqualified,
        )

    def _help_test_pmod_simple(self, statement):
        result = self.parser.parseString(statement)

        expected = {
            KIND: PMOD,
            CONCEPT: {
                NAMESPACE: "go",
                NAME: "protein phosphorylation",
                IDENTIFIER: "0006468",
            },
        }
        self.assertEqual(expected, ProteinModification("Ph"))
        self.assertEqual(expected, result.asDict())

    def test_bel_name(self):
        # long function, legacy modification
        self._help_test_pmod_simple("proteinModification(P)")
        # long function, new modification
        self._help_test_pmod_simple("proteinModification(Ph)")

        # short function, legacy modification
        self._help_test_pmod_simple("pmod(P)")
        # short function, new modification
        self._help_test_pmod_simple("pmod(Ph)")

    def _help_test_pmod_with_residue(self, statement):
        result = self.parser.parseString(statement)

        expected = {
            KIND: PMOD,
            CONCEPT: {
                NAMESPACE: "go",
                NAME: "protein phosphorylation",
                IDENTIFIER: "0006468",
            },
            PMOD_CODE: "Ser",
        }
        self.assertEqual(expected, ProteinModification("Ph", code="Ser"))
        self.assertEqual(expected, result.asDict())

    def test_residue(self):
        # short amino acid
        self._help_test_pmod_with_residue("pmod(Ph, S)")
        # long amino acid
        self._help_test_pmod_with_residue("pmod(Ph, Ser)")

    def _help_test_pmod_full(self, statement):
        result = self.parser.parseString(statement)

        expected = {
            KIND: PMOD,
            CONCEPT: {
                NAMESPACE: "go",
                NAME: "protein phosphorylation",
                IDENTIFIER: "0006468",
            },
            PMOD_CODE: "Ser",
            PMOD_POSITION: 473,
        }

        self.assertEqual(expected, ProteinModification("Ph", code="Ser", position=473))
        self.assertEqual(expected, result.asDict())

    def test_full(self):
        self._help_test_pmod_full("proteinModification(P, Ser, 473)")
        self._help_test_pmod_full("proteinModification(P, S, 473)")
        self._help_test_pmod_full("proteinModification(Ph, Ser, 473)")
        self._help_test_pmod_full("proteinModification(Ph, S, 473)")
        self._help_test_pmod_full("pmod(P, Ser, 473)")
        self._help_test_pmod_full("pmod(P, S, 473)")
        self._help_test_pmod_full("pmod(Ph, Ser, 473)")
        self._help_test_pmod_full("pmod(Ph, S, 473)")

    def _help_test_non_standard_namespace(self, statement):
        result = self.parser.parseString(statement)

        expected = {
            KIND: PMOD,
            CONCEPT: Entity(namespace="MOD", name="PhosRes"),
            PMOD_CODE: "Ser",
            PMOD_POSITION: 473,
        }

        self.assertEqual(
            expected,
            ProteinModification(name="PhosRes", namespace="MOD", code="Ser", position=473),
        )
        self.assertEqual(expected, result.asDict())

    def test_full_with_non_standard_namespace(self):
        self._help_test_non_standard_namespace("proteinModification(MOD:PhosRes, S, 473)")
        self._help_test_non_standard_namespace("proteinModification(MOD:PhosRes, Ser, 473)")
        self._help_test_non_standard_namespace("proteinModification(MOD:PhosRes, S, 473)")
        self._help_test_non_standard_namespace("proteinModification(MOD:PhosRes, Ser, 473)")


class TestGeneModification(unittest.TestCase):
    def setUp(self):
        identifier_parser = ConceptParser()
        self.parser = get_gene_modification_language(
            concept_fqualified=identifier_parser.identifier_fqualified,
            concept_qualified=identifier_parser.identifier_qualified,
        )

        self.expected = GeneModification("Me")

    def test_dsl(self):
        self.assertEqual(
            {
                KIND: GMOD,
                CONCEPT: {
                    NAME: "DNA methylation",
                    IDENTIFIER: "0006306",
                    NAMESPACE: "go",
                },
            },
            self.expected,
        )

    def test_gmod_short(self):
        statement = "geneModification(M)"
        result = self.parser.parseString(statement)
        self.assertEqual(self.expected, result.asDict())

    def test_gmod_unabbreviated(self):
        statement = "geneModification(Me)"
        result = self.parser.parseString(statement)
        self.assertEqual(self.expected, result.asDict())

    def test_gmod_long(self):
        statement = "geneModification(methylation)"
        result = self.parser.parseString(statement)
        self.assertEqual(self.expected, result.asDict())


class TestProteinSubstitution(unittest.TestCase):
    def setUp(self):
        self.parser = get_protein_substitution_language()

    def test_psub_1(self):
        statement = "sub(A, 127, Y)"
        result = self.parser.parseString(statement)

        expected_list = Hgvs("p.Ala127Tyr")
        self.assertEqual(expected_list, result.asDict())

    def test_psub_2(self):
        statement = "sub(Ala, 127, Tyr)"
        result = self.parser.parseString(statement)

        expected_list = Hgvs("p.Ala127Tyr")
        self.assertEqual(expected_list, result.asDict())


class TestGeneSubstitutionParser(unittest.TestCase):
    def setUp(self):
        self.parser = get_gene_substitution_language()

    def test_gsub(self):
        statement = "sub(G,308,A)"
        result = self.parser.parseString(statement)

        expected_dict = Hgvs("c.308G>A")
        self.assertEqual(expected_dict, result.asDict())


class TestFragmentParser(unittest.TestCase):
    """See http://openbel.org/language/web/version_2.0/bel_specification_version_2.0.html#_examples_2"""

    def setUp(self):
        self.parser = get_fragment_language()

    def _help_test_known_length(self, s):
        result = self.parser.parseString(s)
        expected = {KIND: FRAGMENT, FRAGMENT_START: 5, FRAGMENT_STOP: 20}
        self.assertEqual(expected, result.asDict())

    def test_known_length_unquoted(self):
        """test known length"""
        s = "frag(5_20)"
        self._help_test_known_length(s)

    def test_known_length_quotes(self):
        """test known length"""
        s = 'frag("5_20")'
        self._help_test_known_length(s)

    def _help_test_unknown_length(self, s):
        result = self.parser.parseString(s)
        expected = {KIND: FRAGMENT, FRAGMENT_START: 1, FRAGMENT_STOP: "?"}
        self.assertEqual(expected, result.asDict())

    def test_unknown_length_unquoted(self):
        """amino-terminal fragment of unknown length"""
        s = "frag(1_?)"
        self._help_test_unknown_length(s)

    def test_unknown_length_quoted(self):
        """amino-terminal fragment of unknown length"""
        s = 'frag("1_?")'
        self._help_test_unknown_length(s)

    def _help_test_unknown_start_stop(self, s):
        result = self.parser.parseString(s)
        expected = {KIND: FRAGMENT, FRAGMENT_START: "?", FRAGMENT_STOP: "*"}
        self.assertEqual(expected, result.asDict())

    def test_unknown_start_stop_unquoted(self):
        """fragment with unknown start/stop"""
        s = "frag(?_*)"
        self._help_test_unknown_start_stop(s)

    def test_unknown_start_stop_quoted(self):
        """fragment with unknown start/stop"""
        s = 'frag("?_*")'
        self._help_test_unknown_start_stop(s)

    def _help_test_descriptor(self, s):
        result = self.parser.parseString(s)
        expected = {KIND: FRAGMENT, FRAGMENT_MISSING: "?", FRAGMENT_DESCRIPTION: "55kD"}
        self.assertEqual(expected, result.asDict())

    def test_descriptor_unquoted(self):
        """fragment with unknown start/stop and a descriptor"""
        s = 'frag(?, "55kD")'
        self._help_test_descriptor(s)

    def test_descriptor_quoted(self):
        """fragment with unknown start/stop and a descriptor"""
        s = 'frag("?", "55kD")'
        self._help_test_descriptor(s)


class TestTruncationParser(unittest.TestCase):
    def setUp(self):
        self.parser = get_truncation_language()

    def test_trunc_1(self):
        statement = "trunc(40)"
        result = self.parser.parseString(statement)

        expected = Hgvs("p.40*")
        self.assertEqual(expected, result.asDict())

    def test_trunc_2(self):
        """Test a truncation in which the amino acid is specified."""
        statement = "trunc(Gly40)"
        result = self.parser.parseString(statement)

        expected = Hgvs("p.Gly40*")
        self.assertEqual(expected, result.asDict())

    def test_trunc_missing_number(self):
        """Test that an error is raised for a truncation in which the position is omitted."""
        statement = "trunc(Gly)"
        with self.assertRaises(Exception):
            self.parser.parseString(statement)


class TestFusionParser(unittest.TestCase):
    def setUp(self):
        identifier_parser = ConceptParser(namespace_to_pattern={"HGNC": re.compile(".*")})
        identifier_qualified = identifier_parser.identifier_qualified
        self.parser = get_fusion_language(identifier_qualified)

    def test_rna_fusion_known_breakpoints(self):
        """RNA abundance of fusion with known breakpoints"""
        statement = "fus(HGNC:TMPRSS2, r.1_79, HGNC:ERG, r.312_5034)"
        result = self.parser.parseString(statement)

        expected = {
            PARTNER_5P: {
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "TMPRSS2",
                },
            },
            RANGE_5P: {FUSION_REFERENCE: "r", FUSION_START: 1, FUSION_STOP: 79},
            PARTNER_3P: {
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "ERG",
                },
            },
            RANGE_3P: {
                FUSION_REFERENCE: "r",
                FUSION_START: 312,
                FUSION_STOP: 5034,
            },
        }

        self.assertEqual(expected, result.asDict())

    def test_rna_fusion_unspecified_breakpoints(self):
        """RNA abundance of fusion with unspecified breakpoints"""
        statement = "fus(HGNC:TMPRSS2, ?, HGNC:ERG, ?)"
        result = self.parser.parseString(statement)

        expected = {
            PARTNER_5P: {
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "TMPRSS2",
                }
            },
            RANGE_5P: {FUSION_MISSING: "?"},
            PARTNER_3P: {
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "ERG",
                },
            },
            RANGE_3P: {FUSION_MISSING: "?"},
        }

        self.assertEqual(expected, result.asDict())

    def test_rna_fusion_specified_one_fuzzy_breakpoint(self):
        """RNA abundance of fusion with unspecified breakpoints"""
        statement = "fusion(HGNC:TMPRSS2, r.1_79, HGNC:ERG, r.?_1)"
        result = self.parser.parseString(statement)

        expected = {
            PARTNER_5P: {
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "TMPRSS2",
                },
            },
            RANGE_5P: {FUSION_REFERENCE: "r", FUSION_START: 1, FUSION_STOP: 79},
            PARTNER_3P: {
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "ERG",
                },
            },
            RANGE_3P: {FUSION_REFERENCE: "r", FUSION_START: "?", FUSION_STOP: 1},
        }

        self.assertEqual(expected, result.asDict())

    def test_rna_fusion_specified_fuzzy_breakpoints(self):
        """RNA abundance of fusion with unspecified breakpoints"""
        statement = "fusion(HGNC:TMPRSS2, r.1_?, HGNC:ERG, r.?_1)"
        result = self.parser.parseString(statement)

        expected = {
            PARTNER_5P: {
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "TMPRSS2",
                },
            },
            RANGE_5P: {FUSION_REFERENCE: "r", FUSION_START: 1, FUSION_STOP: "?"},
            PARTNER_3P: {
                CONCEPT: {
                    NAMESPACE: "HGNC",
                    NAME: "ERG",
                },
            },
            RANGE_3P: {FUSION_REFERENCE: "r", FUSION_START: "?", FUSION_STOP: 1},
        }

        self.assertEqual(expected, result.asDict())


class TestLocation(unittest.TestCase):
    def setUp(self):
        identifier_parser = ConceptParser(namespace_to_pattern={"GO": re.compile(".*")})
        identifier_qualified = identifier_parser.identifier_qualified
        self.parser = get_location_language(identifier_qualified)

    def test_a(self):
        statement = "loc(GO:intracellular)"
        result = self.parser.parseString(statement)
        expected = {LOCATION: {NAMESPACE: "GO", NAME: "intracellular"}}
        self.assertEqual(expected, result.asDict())



# -*- coding: utf-8 -*-

"""Summary functions for citation and provenance information in BEL graphs."""

from typing import Iterable, Set

from ..filters.edge_predicates import CITATION_PREDICATES
from ..graph import BELGraph
from ...constants import CITATION, IDENTIFIER

__all__ = [
    "iterate_pubmed_identifiers",
    "iterate_pmc_identifiers",
    "get_pubmed_identifiers",
    "get_pmc_identifiers",
]


def iterate_citation_identifiers(graph, prefix: str):
    """Iterate over all citation identifiers with the given prefix in a graph.

    :param graph: A BEL graph
    :param prefix: The citation prefix to keep
    :return: An iterator over the PubMed identifiers in the graph
    """
    predicate = CITATION_PREDICATES.get(prefix)
    if predicate is None:
        raise ValueError(f"Invalid citation prefix: {prefix}")

    return (data[CITATION][IDENTIFIER].strip() for _, _, data in graph.edges(data=True) if predicate(data))


def iterate_pubmed_identifiers(graph: BELGraph) -> Iterable[str]:
    """Iterate over all PubMed identifiers in a graph.

    :param graph: A BEL graph
    :return: An iterator over the PubMed identifiers in the graph
    """
    return iterate_citation_identifiers(graph, "pubmed")


def iterate_pmc_identifiers(graph: BELGraph) -> Iterable[str]:
    """Iterate over all PMC identifiers in a graph.

    :param graph: A BEL graph
    :return: An iterator over the PMC identifiers in the graph
    """
    return iterate_citation_identifiers(graph, "pmc")


def get_citation_identifiers(graph: BELGraph, prefix: str) -> Set[str]:
    """Get the set of all identifiers with the give prefix cited in the construction of a graph.

    :param graph: A BEL graph
    :param prefix: The citation prefix to keep
    :return: A set of all PubMed identifiers cited in the construction of this graph
    """
    return set(iterate_citation_identifiers(graph, prefix))


def get_pubmed_identifiers(graph: BELGraph) -> Set[str]:
    """Get the set of all PubMed identifiers cited in the construction of a graph.

    :param graph: A BEL graph
    :return: A set of all PubMed identifiers cited in the construction of this graph
    """
    return get_citation_identifiers(graph, "pubmed")


def get_pmc_identifiers(graph: BELGraph) -> Set[str]:
    """Get the set of all PMC identifiers cited in the construction of a graph.

    :param graph: A BEL graph
    :return: A set of all PMC identifiers cited in the construction of this graph
    """
    return get_citation_identifiers(graph, "pmc")
